/*
 * ----------------------------------------------
 *
 * This software belongs to the zzzutils package.
 *
 * Copyright (c) 2003 Fabian "zzznowman" Pietsch.
 * All rights reserved.
 *
 * ----------------------------------------------
 */

/*
 * zzzupdate-xmms.c -- search filesystem for audio files & enqueue them in xmms
 *
 ******************************************************************************
 *
 * brief ChangeLog:
 *   2003-05-05 original implementation as shell script
 *   2003-08-06 major changes & enhancements
 *   2003-08-06 reimplementation in C
 *
 ******************************************************************************
 *
 *
 */

#include "zzzutils.h"
#include <stdlib.h>
#include <stdio.h>

/* required for stat() */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

/* required for readdir() */
#include <dirent.h>

/* required for strcoll() */
#include <string.h>


struct zbtree {
	char *v;		/* value		*/
	struct zbtree *n_le;	/* next, less or equal	*/
	struct zbtree *n_gt;	/* next, greater	*/
};

struct zbtree_enum {
	int stack_depth;
	int stack_size;
	struct zbtree **stack;
};
#define ZBTREE_SINC 4

#if defined(DEBUG)
/* zbtree_dumpenum -- zbtree_enum corruption tracking */
void zbtree_dumpenum(struct zbtree_enum *e)
{
	int i;
	
	zwarn("zbtree_dumpenum(): ===begin=== for enum at %p", e);
	zwarn("zbtree_dumpenum(): stack_depth == %i, stack_size == %i", e->stack_depth, e->stack_size);

	/* check for general stack validness */
	if (e->stack_depth > e->stack_size)
		zerror("zbtree_dumpenum(): !!! stack_depth > stack_size !!!");
	else {
		zwarn("zbtree_dumpenum(): stack dump:");
		/* loop through e's stack */
		for (i = 0; i < e->stack_depth; i++)
		{
			zwarn("zbtree_dumpenum(): e->stack[%i] == %p", i, e->stack[i]);
		}
	}
	
	zwarn("zbtree_dumpenum(): === end === for enum at %p", e);
}
#endif

/* zbtree_add -- add a new element to a struct zbtree chain */
struct zbtree *zbtree_add(struct zbtree **bt, char *v)
{
	struct zbtree  *bt_new,
	              **bt_cur = bt;
	
	/* create new zbtree element */
	if ((bt_new = (struct zbtree *) calloc(1, sizeof(struct zbtree))) == NULL) {
		zperror("zbtree_add(): malloc()");
		return NULL;
	}
	bt_new->v = v;

	/* search free hook in bt to link bt_new into */
	while (*bt_cur != NULL) {
		/* compare new element's value to that of the current element's */
		if (strcoll(v, (**bt_cur).v) > 0)
			bt_cur = &(**bt_cur).n_gt;
		else
			bt_cur = &(**bt_cur).n_le;
	}

	/* link bt_new into free hook in bt */
	*bt_cur = bt_new;
	
	/* return pointer to added element */
	return bt_new;
}

/* zbtree_free -- free an entire struct zbtree chain */
void zbtree_free(struct zbtree **bt, const int freev)
{
	/* subtress to free? */
	if ((**bt).n_gt != NULL)
		zbtree_free(&(**bt).n_gt, freev);
	if ((**bt).n_le != NULL)
		zbtree_free(&(**bt).n_le, freev);

	/* TODO: remove this---> */
	/* free(**bt);
	*bt = NULL; */
	/* <--- */

	/* also free contained data? */
	if (freev)
		(**bt).v = realloc((**bt).v, 0);

	/* finally free zbtree element itself */
	*bt = realloc(*bt, 0);
}

/* zbtree_next -- allow enumeration through zbtree elements
 * Attention: e->stack _must_ be NULL if zbtree_next(..., e) never called before!
 */
struct zbtree *zbtree_next(struct zbtree *bt, struct zbtree_enum *e)
{
	struct zbtree *bt_cur, *bt_next;
	
	/* begin of enumeration? */
	if ((e->stack == NULL) || (e->stack[0] != bt)) {
#ifdef DEBUG
		zwarn("zbtree_next(): new enumeration");
#endif
		if (e->stack == NULL) {
#ifdef DEBUG
			zwarn("zbtree_next(): new enum: e->stack == NULL");
#endif
			e->stack_size = 0;
		}
		e->stack_depth = 0;

		/* first call, so start least-element search on root element */
		bt_next = bt;
	}
	else {
		bt_next = NULL;

		/* dequeue elements on stack until one of the following occurs:
		 *   - element's value has not been returned yet
		 *     (element has been enqueued during least-element search)
		 *   - element has a n_gt subtree which has not been processed yet
		 *     (<=> subtree's address != priviously dequeued element's)
		 *   - stack gets empty
		 *     (enumeration is completed, then)
		 */
		do {
#ifdef DEBUG
			zwarn("zbtree_next(): dequeueing loop");
#endif

			/* stack size shrinkable by more than ZBTREE_SINC elements? */
			if ((e->stack_size - e->stack_depth) > ZBTREE_SINC) {
#ifdef DEBUG
				zwarn("zbtree_next(): dequeueing loop: realloc()ing stack (%p)", e->stack);
#endif
				e->stack_size -= ZBTREE_SINC;
				if ((e->stack = realloc(e->stack,
				                        e->stack_size
							  * sizeof(struct zbtree *)))
				    == NULL) {
					zperror("zbtree_next(): malloc()");
					return NULL;
				}
			}

			/* remember last element dequeued */
			bt_cur = bt_next;
			
			/* examine last element on stack */
			bt_next = e->stack[e->stack_depth - 1];
#ifdef DEBUG
			zwarn("zbtree_next(): dequeueing loop: stack indexing ok: bt_next == %p", bt_next);
#endif

			/* returning from branch to lesser element?
			 * => element itself has not been returned yet
			 */
			if (bt_next->n_le && (bt_next->n_le == bt_cur))
				return bt_next;
			
			/* continue loop (see above)
			 ***************************
			 * dequeue last element on stack
			 * if previous loop continuation conditions satisfied
			 */
		} while ((!bt_next->n_gt || (bt_next->n_gt == bt_cur)) &&
		         (--(e->stack_depth) > 0));
		
		/* last element reached? */
		if (e->stack_depth == 0) {
#ifdef DEBUG
			zwarn("zbtree_next(): dequeueing loop: last element reached");
#endif
			/* do cleanup: deallocate stack */
			e->stack = realloc(e->stack, 0);	
			e->stack_size = 0;
			
			/* return NULL to indicate end of enumeration */
			return NULL;
		}

		/* branch to previously unprocessed greater-than subtree */
		bt_next = bt_next->n_gt;
	}

	/* do { ... } until there's no n_le / lesser value */
	do {
#ifdef DEBUG
		zwarn("zbtree_next(): enqueueing loop");
#endif
		
		/* continue with next element (see above & below) */
		bt_cur = bt_next;
		
		/* need to malloc() bigger stack? */
		if (e->stack_depth == e->stack_size) {
#ifdef DEBUG
			zwarn("zbtree_next(): changing realloc()ing stack (%p)", e->stack);
#endif
			e->stack_size += ZBTREE_SINC;
			if ((e->stack = (struct zbtree **)
			                  realloc(e->stack,
				                  e->stack_size
						    * sizeof(struct zbtree *)))
			    == NULL) {
				zperror("zbtree_next(): malloc()");
				return NULL;
			}
		}
		
		/* put bt_cur on the stack
		 *************************
		 * Each element gets put onto the stack at least once:
		 * when it's value is returned. After that,
		 * there's still it's n_gt... (being processed by next call)
		 */
#ifdef DEBUG
		zwarn("zbtree_next(): enqueueing loop: putting (bt_cur == %p) onto the stack at %i", bt_cur, e->stack_depth);
#endif
		e->stack[e->stack_depth++] = bt_cur;

		/* continue with next element lesser than this one */
		bt_next = bt_cur->n_le;
	} while (bt_next);  /* stop when subtree's least element found */

	/* bt_cur now is on the stack and has no lesser element */
	return bt_cur;
}


/* init dirs count (cf. argv) */
int initdirsc = 1;

/* init dirs values (cf. argv) */
char *initdirsv[] = {
	"sto"
};

/* pointer to array of initial dirs
 * if program arguments present, points to first directory argument
 */
char **initdirsvp = initdirsv;


/* search directory -- recursively calls itself with subdirectories */
int searchdir(const char *name)
{
	struct stat dir_info,		/* stat() for name,		*/
	            ent_info;		/*            dir entry		*/
	DIR *dir;			/* pointer to directory stream	*/
	struct dirent *dirent;		/* directory entry		*/
	struct zbtree *dir_dirs=NULL,	/* directory's subdirs,		*/
	              *dir_files=NULL,	/*             files		*/
		      *bt_cur;		/* zbtree enumeration		*/
	struct zbtree_enum zbte;	/* enumerator for zbtree_next	*/
		      
	/* stat name (for presence & directory check) */
	if (stat(name, &dir_info) != 0) {
		zperror("can't stat %s", name);
		return -1;
	}
	
	/* is name a directory? */
	if (!S_ISDIR(dir_info.st_mode)) {
		zerror("%s is not a dir", name);
		return -1;
	}
	printf("scanning %s...\n", name);

	/* open directory stream */
	if ((dir = opendir(name)) == NULL) {
		zperror("opendir(\"%s\")", name);
		return -1;
	}

	/* enter directory */
	if (chdir(name) != 0) {
		zperror("couldn't chdir() to %s", name);
		closedir(dir);
		return -1;
	}
	
	/* loop through directory entries */
	while ((dirent = readdir(dir))) {
		/* stat directory entry */
		if (stat(dirent->d_name, &ent_info) != 0) {
			zperror("can't stat %s", dirent->d_name);
			continue;
		}
		
		if (S_ISDIR(ent_info.st_mode)) {
			zbtree_add(&dir_dirs, strdup(dirent->d_name));
			zwarn("readdir(\"%s\"): %s is a directory", name, (*dirent).d_name);
		}
		else {
			zbtree_add(&dir_files, strdup(dirent->d_name));
			zwarn("readdir(\"%s\"): %s is not a directory", name, (*dirent).d_name);
		}
	}
	dirent = NULL;

	/* process gathered information */
	printf("\n\nSummary for %s:\n\n", name);

	/* any directories found? */
	if (dir_dirs) {
		printf("Directories:\n");
		zbte.stack = NULL;
		while ((bt_cur = zbtree_next(dir_dirs, &zbte)))
			printf("  %s\n", bt_cur->v);
		printf("\n");
	}

	/* any files found? */
	if (dir_files) {
		printf("Files:\n");
		while ((bt_cur = zbtree_next(dir_files, &zbte)))
			printf("  %s\n", bt_cur->v);
		printf("\n");
	}

	/* free used zbtrees
	 * (including d_name string copies malloc()ed by strdup())
	 */
	zbtree_free(&dir_dirs, 1);
	zbtree_free(&dir_files, 1);
	
	/* leave directory */
	if (chdir("..") != 0) {
		closedir(dir);
		zperror("couldn't chdir() from %s to parent directory", name);
		return -2;
	}
	
	/* close directory stream */
	closedir(dir);
	return 0;
}

int main(const int argc, char *argv[])
{
	int initdir_idx;
	
	if (argc > 1) {
		initdirsvp = &argv[1];
		initdirsc = argc - 1;
	}
	for (initdir_idx = 0; initdir_idx < initdirsc; initdir_idx++) {
		searchdir(initdirsvp[initdir_idx]);
	}
	return 0;
}
