/*
 * ----------------------------------------------
 *
 * This software belongs to the zzzutils package.
 *
 * Copyright (c) 2003 Fabian "zzznowman" Pietsch.
 * All rights reserved.
 *
 * ----------------------------------------------
 */

/*
 * zzzupdate-xmms.c -- search filesystem for audio files & enqueue them in xmms
 *
 ******************************************************************************
 *
 * brief ChangeLog:
 *   2003-05-05 original implementation as shell script
 *   2003-08-06 major changes & enhancements
 *   2003-08-06 reimplementation in C
 *   2003-08-08 finished zbtree implementation - now in zbtree.{c,h}
 *
 ******************************************************************************
 *
 *
 */

#include "zzzutils.h"
#include "zbtree.h"
#include <stdio.h>
#include <string.h>

/* required for stat() */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

/* required for readdir() */
#include <dirent.h>


/* init dirs count (cf. argv) */
int initdirsc = 1;

/* init dirs values (cf. argv) */
char *initdirsv[] = {
	"sto"
};

/* pointer to array of initial dirs
 * if program arguments present, points to first directory argument
 */
char **initdirsvp = initdirsv;


/* search directory -- recursively calls itself with subdirectories */
int searchdir(const char *name)
{
	struct stat dir_info,		/* stat() for name,		*/
	            ent_info;		/*            dir entry		*/
	DIR *dir;			/* pointer to directory stream	*/
	struct dirent *dirent;		/* directory entry		*/
	struct zbtree *dir_dirs=NULL,	/* directory's subdirs,		*/
	              *dir_files=NULL,	/*             files		*/
		      *bt_cur;		/* zbtree enumeration		*/
	struct zbtree_enum zbte;	/* enumerator for zbtree_next	*/
		      
	/* stat name (for presence & directory check) */
	if (stat(name, &dir_info) != 0) {
		zperror("can't stat %s", name);
		return -1;
	}
	
	/* is name a directory? */
	if (!S_ISDIR(dir_info.st_mode)) {
		zerror("%s is not a dir", name);
		return -1;
	}
	printf("scanning %s...\n", name);

	/* open directory stream */
	if ((dir = opendir(name)) == NULL) {
		zperror("opendir(\"%s\")", name);
		return -1;
	}

	/* enter directory */
	if (chdir(name) != 0) {
		zperror("couldn't chdir() to %s", name);
		closedir(dir);
		return -1;
	}
	
	/* loop through directory entries */
	while ((dirent = readdir(dir))) {
		/* "." or ".."? */
		if (dirent->d_name[0] == '.') {
			switch (dirent->d_name[1]) {
			case 0:
			case '.':
#ifdef DEBUG
				zwarn("rejecting %s", dirent->d_name);
#endif
				continue;
			}
		}
		
		/* stat directory entry */
		if (stat(dirent->d_name, &ent_info) != 0) {
			zperror("can't stat %s", dirent->d_name);
			continue;
		}
		
		if (S_ISDIR(ent_info.st_mode)) {
			zbtree_add(&dir_dirs, strdup(dirent->d_name));
#ifdef DEBUG
			zwarn("readdir(\"%s\"): %s is a directory", name, dirent->d_name);
#endif
		}
		else {
			zbtree_add(&dir_files, strdup(dirent->d_name));
#ifdef DEBUG
			zwarn("readdir(\"%s\"): %s is not a directory", name, dirent->d_name);
#endif
		}
	}
	dirent = NULL;

	/* process gathered information */
	printf("\n\nSummary for %s:\n\n", name);

	/* any directories found? */
	if (dir_dirs) {
		printf("Directories:\n");

		/* currently zbtree_enum.stack _must_ be NULL
		 * when zbtree_next() first called
		 */
		zbte.stack = NULL;

		while ((bt_cur = zbtree_next(dir_dirs, &zbte)))
			printf("  %s\n", bt_cur->v);
		printf("\n");
	}

	/* any files found? */
	if (dir_files) {
		printf("Files:\n");

		/* currently zbtree_enum.stack _must_ be NULL
		 * when zbtree_next() first called
		 */
		zbte.stack = NULL;

		while ((bt_cur = zbtree_next(dir_files, &zbte)))
			printf("  %s\n", bt_cur->v);
		printf("\n");
	}

	/* free used zbtrees
	 * (including dirent.d_name string copies malloc()ed by strdup())
	 */
	zbtree_free(&dir_dirs, 1);
	zbtree_free(&dir_files, 1);
	
	/* leave directory */
	if (chdir("..") != 0) {
		closedir(dir);
		zperror("couldn't chdir() from %s to parent directory", name);
		return -2;
	}
	
	/* close directory stream */
	closedir(dir);
	return 0;
}

int main(const int argc, char *argv[])
{
	int initdir_idx;
	
	if (argc > 1) {
		initdirsvp = &argv[1];
		initdirsc = argc - 1;
	}
	for (initdir_idx = 0; initdir_idx < initdirsc; initdir_idx++) {
		if (searchdir(initdirsvp[initdir_idx]) < -1) {
			zerror("searchdir() failed in a non-recoverable way -- exiting");
			exit(2);
		}
	}
	exit(0);
	/* return 0; */
}
